@@ -22,11 +22,14 @@ See :doc:`/central_scheduler` for more info.
 """
 
 import collections
+import inspect
+
 try:
     import cPickle as pickle
 except ImportError:
     import pickle
 import functools
+import hashlib
 import itertools
 import logging
 import os
@@ -39,21 +42,11 @@ from luigi import configuration
 from luigi import notifications
 from luigi import parameter
 from luigi import task_history as history
-from luigi.task_status import DISABLED, DONE, FAILED, PENDING, RUNNING, SUSPENDED, UNKNOWN
+from luigi.task_status import DISABLED, DONE, FAILED, PENDING, RUNNING, SUSPENDED, UNKNOWN, \
+    BATCH_RUNNING
 from luigi.task import Config
 
-logger = logging.getLogger("luigi.server")
-
-
-class Scheduler(object):
-    """
-    Abstract base class.
-
-    Note that the methods all take string arguments, not Task objects...
-    """""
-    add_task = NotImplemented
-    get_work = NotImplemented
-    ping = NotImplemented
+logger = logging.getLogger(__name__)
 
 UPSTREAM_RUNNING = 'UPSTREAM_RUNNING'
 UPSTREAM_MISSING_INPUT = 'UPSTREAM_MISSING_INPUT'
@@ -77,6 +70,48 @@ STATUS_TO_UPSTREAM_MAP = {
 
 TASK_FAMILY_RE = re.compile(r'([^(_]+)[(_]')
 
+RPC_METHODS = {}
+
+_retry_policy_fields = [
+    "retry_count",
+    "disable_hard_timeout",
+    "disable_window",
+]
+RetryPolicy = collections.namedtuple("RetryPolicy", _retry_policy_fields)
+
+
+def _get_empty_retry_policy():
+    return RetryPolicy(*[None] * len(_retry_policy_fields))
+
+
+def rpc_method(**request_args):
+    def _rpc_method(fn):
+        # If request args are passed, return this function again for use as
+        # the decorator function with the request args attached.
+        fn_args = inspect.getargspec(fn)
+
+        assert not fn_args.varargs
+        assert fn_args.args[0] == 'self'
+        all_args = fn_args.args[1:]
+        defaults = dict(zip(reversed(all_args), reversed(fn_args.defaults or ())))
+        required_args = frozenset(arg for arg in all_args if arg not in defaults)
+        fn_name = fn.__name__
+
+        @functools.wraps(fn)
+        def rpc_func(self, *args, **kwargs):
+            actual_args = defaults.copy()
+            actual_args.update(dict(zip(all_args, args)))
+            actual_args.update(kwargs)
+            if not all(arg in actual_args for arg in required_args):
+                raise TypeError('{} takes {} arguments ({} given)'.format(
+                    fn_name, len(all_args), len(actual_args)))
+            return self._request('/api/{}'.format(fn_name), actual_args, **request_args)
+
+        RPC_METHODS[fn_name] = rpc_func
+        return fn
+
+    return _rpc_method
+
 
 class scheduler(Config):
     # TODO(erikbern): the config_path is needed for backwards compatilibity. We
@@ -86,24 +121,26 @@ class scheduler(Config):
     worker_disconnect_delay = parameter.FloatParameter(default=60.0)
     state_path = parameter.Parameter(default='/var/lib/luigi-server/state.pickle')
 
-    # Jobs are disabled if we see more than disable_failures failures in disable_window seconds.
+    # Jobs are disabled if we see more than retry_count failures in disable_window seconds.
     # These disables last for disable_persist seconds.
     disable_window = parameter.IntParameter(default=3600,
                                             config_path=dict(section='scheduler', name='disable-window-seconds'))
-    disable_failures = parameter.IntParameter(default=None,
-                                              config_path=dict(section='scheduler', name='disable-num-failures'))
-    disable_hard_timeout = parameter.IntParameter(default=None,
+    retry_count = parameter.IntParameter(default=999999999,
+                                         config_path=dict(section='scheduler', name='disable_failures'))
+    disable_hard_timeout = parameter.IntParameter(default=999999999,
                                                   config_path=dict(section='scheduler', name='disable-hard-timeout'))
     disable_persist = parameter.IntParameter(default=86400,
                                              config_path=dict(section='scheduler', name='disable-persist-seconds'))
     max_shown_tasks = parameter.IntParameter(default=100000)
     max_graph_nodes = parameter.IntParameter(default=100000)
-    prune_done_tasks = parameter.BoolParameter(default=False)
 
     record_task_history = parameter.BoolParameter(default=False)
 
     prune_on_get_work = parameter.BoolParameter(default=False)
 
+    def _get_retry_policy(self):
+        return RetryPolicy(self.retry_count, self.disable_hard_timeout, self.disable_window)
+
 
 class Failures(object):
     """
@@ -160,10 +197,8 @@ def _get_default(x, default):
 
 
 class Task(object):
-
     def __init__(self, task_id, status, deps, resources=None, priority=0, family='', module=None,
-                 params=None, disable_failures=None, disable_window=None, disable_hard_timeout=None,
-                 tracking_url=None):
+                 params=None, tracking_url=None, status_message=None, retry_policy='notoptional'):
         self.id = task_id
         self.stakeholders = set()  # workers ids that are somehow related to this task (i.e. don't prune while any of these workers are still active)
         self.workers = set()  # workers ids that can perform task - task is 'BROKEN' if none of these workers are active
@@ -184,35 +219,42 @@ class Task(object):
         self.family = family
         self.module = module
         self.params = _get_default(params, {})
-        self.disable_failures = disable_failures
-        self.disable_hard_timeout = disable_hard_timeout
-        self.failures = Failures(disable_window)
+
+        self.retry_policy = retry_policy
+        self.failures = Failures(self.retry_policy.disable_window)
         self.tracking_url = tracking_url
+        self.status_message = status_message
         self.scheduler_disable_time = None
         self.runnable = False
+        self.batchable = False
+        self.batch_id = None
 
     def __repr__(self):
         return "Task(%r)" % vars(self)
 
+    # TODO(2017-08-10) replace this function with direct calls to batchable
+    # this only exists for backward compatibility
+    def is_batchable(self):
+        try:
+            return self.batchable
+        except AttributeError:
+            return False
+
     def add_failure(self):
         self.failures.add_failure()
 
     def has_excessive_failures(self):
-        if (self.failures.first_failure_time is not None and
-                self.disable_hard_timeout):
-            if (time.time() >= self.failures.first_failure_time +
-                    self.disable_hard_timeout):
+        if self.failures.first_failure_time is not None:
+            if (time.time() >= self.failures.first_failure_time + self.retry_policy.disable_hard_timeout):
                 return True
 
-        if self.failures.num_failures() >= self.disable_failures:
+        logger.debug('%s task num failures is %s and limit is %s', self.id, self.failures.num_failures(), self.retry_policy.retry_count)
+        if self.failures.num_failures() >= self.retry_policy.retry_count:
+            logger.debug('%s task num failures limit(%s) is exceeded', self.id, self.retry_policy.retry_count)
             return True
 
         return False
 
-    def can_disable(self):
-        return (self.disable_failures is not None or
-                self.disable_hard_timeout is not None)
-
     @property
     def pretty_id(self):
         param_str = ', '.join('{}={}'.format(key, value) for key, value in self.params.items())
@@ -294,12 +336,15 @@ class SimpleTaskState(object):
         self._tasks = {}  # map from id to a Task object
         self._status_tasks = collections.defaultdict(dict)
         self._active_workers = {}  # map from id to a Worker object
+        self._task_batchers = {}
 
     def get_state(self):
-        return self._tasks, self._active_workers
+        return self._tasks, self._active_workers, self._task_batchers
 
     def set_state(self, state):
-        self._tasks, self._active_workers = state
+        self._tasks, self._active_workers = state[:2]
+        if len(state) >= 3:
+            self._task_batchers = state[2]
 
     def dump(self):
         try:
@@ -318,7 +363,7 @@ class SimpleTaskState(object):
                 with open(self._state_path, 'rb') as fobj:
                     state = pickle.load(fobj)
             except BaseException:
-                logger.exception("Error when loading state. Starting from clean slate.")
+                logger.exception("Error when loading state. Starting from empty state.")
                 return
 
             self.set_state(state)
@@ -326,7 +371,7 @@ class SimpleTaskState(object):
             for task in six.itervalues(self._tasks):
                 self._status_tasks[task.status][task.id] = task
         else:
-            logger.info("No prior state file exists at %s. Starting with clean slate", self._state_path)
+            logger.info("No prior state file exists at %s. Starting with empty state", self._state_path)
 
     def get_active_tasks(self, status=None):
         if status:
@@ -336,6 +381,13 @@ class SimpleTaskState(object):
             for task in six.itervalues(self._tasks):
                 yield task
 
+    def get_batch_running_tasks(self, batch_id):
+        assert batch_id is not None
+        return [
+            task for task in self.get_active_tasks(BATCH_RUNNING)
+            if task.batch_id == batch_id
+        ]
+
     def get_running_tasks(self):
         return six.itervalues(self._status_tasks[RUNNING])
 
@@ -343,6 +395,13 @@ class SimpleTaskState(object):
         return itertools.chain.from_iterable(six.itervalues(self._status_tasks[status])
                                              for status in [PENDING, RUNNING])
 
+    def set_batcher(self, worker_id, family, batcher_args, max_batch_size):
+        self._task_batchers.setdefault(worker_id, {})
+        self._task_batchers[worker_id][family] = (batcher_args, max_batch_size)
+
+    def get_batcher(self, worker_id, family):
+        return self._task_batchers.get(worker_id, {}).get(family, (None, 1))
+
     def num_pending_tasks(self):
         """
         Return how many tasks are PENDING + RUNNING. O(1).
@@ -367,13 +426,20 @@ class SimpleTaskState(object):
             self.set_status(task, FAILED, config)
             task.failures.clear()
 
+    def set_batch_running(self, task, batch_id, worker_id):
+        self.set_status(task, BATCH_RUNNING)
+        task.batch_id = batch_id
+        task.worker_running = worker_id
+
     def set_status(self, task, new_status, config=None):
         if new_status == FAILED:
             assert config is not None
 
-        if new_status == DISABLED and task.status == RUNNING:
+        if new_status == DISABLED and task.status in (RUNNING, BATCH_RUNNING):
             return
 
+        remove_on_failure = task.batch_id is not None and not task.batchable
+
         if task.status == DISABLED:
             if new_status == DONE:
                 self.re_enable(task)
@@ -382,7 +448,13 @@ class SimpleTaskState(object):
             elif task.scheduler_disable_time is not None and new_status != DISABLED:
                 return
 
-        if new_status == FAILED and task.can_disable() and task.status != DISABLED:
+        if task.status == RUNNING and task.batch_id is not None:
+            for batch_task in self.get_batch_running_tasks(task.batch_id):
+                self.set_status(batch_task, new_status, config)
+                batch_task.batch_id = None
+            task.batch_id = None
+
+        if new_status == FAILED and task.status != DISABLED:
             task.add_failure()
             if task.has_excessive_failures():
                 task.scheduler_disable_time = time.time()
@@ -391,7 +463,7 @@ class SimpleTaskState(object):
                     'Luigi Scheduler: DISABLED {task} due to excessive failures'.format(task=task.id),
                     '{task} failed {failures} times in the last {window} seconds, so it is being '
                     'disabled for {persist} seconds'.format(
-                        failures=config.disable_failures,
+                        failures=task.retry_policy.retry_count,
                         task=task.id,
                         window=config.disable_window,
                         persist=config.disable_persist,
@@ -405,6 +477,11 @@ class SimpleTaskState(object):
             task.status = new_status
             task.updated = time.time()
 
+        if new_status == FAILED:
+            task.retry = time.time() + config.retry_delay
+            if remove_on_failure:
+                task.remove = time.time()
+
     def fail_dead_worker_task(self, task, config, assistants):
         # If a running worker disconnects, tag all its jobs as FAILED and subject it to the same retry logic
         if task.status == RUNNING and task.worker_running and task.worker_running not in task.stakeholders | assistants:
@@ -417,11 +494,12 @@ class SimpleTaskState(object):
 
     def update_status(self, task, config):
         # Mark tasks with no remaining active stakeholders for deletion
-        if not task.stakeholders:
-            if task.remove is None:
-                logger.info("Task %r has stakeholders %r but none remain connected -> might remove "
-                            "task in %s seconds", task.id, task.stakeholders, config.remove_delay)
-                task.remove = time.time() + config.remove_delay
+        if (not task.stakeholders) and (task.remove is None) and (task.status != RUNNING):
+            # We don't check for the RUNNING case, because that is already handled
+            # by the fail_dead_worker_task function.
+            logger.debug("Task %r has no stakeholders anymore -> might remove "
+                         "task in %s seconds", task.id, config.remove_delay)
+            task.remove = time.time() + config.remove_delay
 
         # Re-enable task after the disable time expires
         if task.status == DISABLED and task.scheduler_disable_time is not None:
@@ -433,7 +511,7 @@ class SimpleTaskState(object):
             self.set_status(task, PENDING, config)
 
     def may_prune(self, task):
-        return task.remove and time.time() > task.remove
+        return task.remove and time.time() >= task.remove
 
     def inactivate_tasks(self, delete_tasks):
         # The terminology is a bit confusing: we used to "delete" tasks when they became inactive,
@@ -449,7 +527,7 @@ class SimpleTaskState(object):
                 continue
             last_get_work = getattr(worker, 'last_get_work', None)
             if last_get_work_gt is not None and (
-                    last_get_work is None or last_get_work <= last_get_work_gt):
+                            last_get_work is None or last_get_work <= last_get_work_gt):
                 continue
             yield worker
 
@@ -479,17 +557,8 @@ class SimpleTaskState(object):
         for worker in workers:
             self.get_worker(worker).disabled = True
 
-    def get_necessary_tasks(self):
-        necessary_tasks = set()
-        for task in self.get_active_tasks():
-            if task.status not in (DONE, DISABLED, UNKNOWN) or \
-                    task.scheduler_disable_time is not None:
-                necessary_tasks.update(task.deps)
-                necessary_tasks.add(task.id)
-        return necessary_tasks
 
-
-class CentralPlannerScheduler(Scheduler):
+class Scheduler(object):
     """
     Async scheduler that can handle multiple workers, etc.
 
@@ -501,7 +570,7 @@ class CentralPlannerScheduler(Scheduler):
         Keyword Arguments:
         :param config: an object of class "scheduler" or None (in which the global instance will be used)
         :param resources: a dict of str->int constraints
-        :param task_history_override: ignore config and use this object as the task history
+        :param task_history_impl: ignore config and use this object as the task history
         """
         self._config = config or scheduler(**kwargs)
         self._state = SimpleTaskState(self._config.state_path)
@@ -514,10 +583,7 @@ class CentralPlannerScheduler(Scheduler):
         else:
             self._task_history = history.NopHistory()
         self._resources = resources or configuration.get_config().getintdict('resources')  # TODO: Can we make this a Parameter?
-        self._make_task = functools.partial(
-            Task, disable_failures=self._config.disable_failures,
-            disable_hard_timeout=self._config.disable_hard_timeout,
-            disable_window=self._config.disable_window)
+        self._make_task = functools.partial(Task, retry_policy=self._config._get_retry_policy())
         self._worker_requests = {}
 
     def load(self):
@@ -526,6 +592,7 @@ class CentralPlannerScheduler(Scheduler):
     def dump(self):
         self._state.dump()
 
+    @rpc_method()
     def prune(self):
         logger.info("Starting pruning of task graph")
         self._prune_workers()
@@ -536,7 +603,7 @@ class CentralPlannerScheduler(Scheduler):
         remove_workers = []
         for worker in self._state.get_active_workers():
             if worker.prune(self._config):
-                logger.info("Worker %s timed out (no contact for >=%ss)", worker, self._config.worker_disconnect_delay)
+                logger.debug("Worker %s timed out (no contact for >=%ss)", worker, self._config.worker_disconnect_delay)
                 remove_workers.append(worker.id)
 
         self._state.inactivate_workers(remove_workers)
@@ -545,15 +612,10 @@ class CentralPlannerScheduler(Scheduler):
         assistant_ids = set(w.id for w in self._state.get_assistants())
         remove_tasks = []
 
-        if assistant_ids:
-            necessary_tasks = self._state.get_necessary_tasks()
-        else:
-            necessary_tasks = ()
-
         for task in self._state.get_active_tasks():
             self._state.fail_dead_worker_task(task, self._config, assistant_ids)
             self._state.update_status(task, self._config)
-            if self._state.may_prune(task) and task.id not in necessary_tasks:
+            if self._state.may_prune(task):
                 logger.info("Removing task %r", task.id)
                 remove_tasks.append(task.id)
 
@@ -580,10 +642,16 @@ class CentralPlannerScheduler(Scheduler):
             if t is not None and prio > t.priority:
                 self._update_priority(t, prio, worker)
 
+    @rpc_method()
+    def add_task_batcher(self, worker, task_family, batched_args, max_batch_size=float('inf')):
+        self._state.set_batcher(worker, task_family, batched_args, max_batch_size)
+
+    @rpc_method()
     def add_task(self, task_id=None, status=PENDING, runnable=True,
                  deps=None, new_deps=None, expl=None, resources=None,
                  priority=0, family='', module=None, params=None,
-                 assistant=False, tracking_url=None, **kwargs):
+                 assistant=False, tracking_url=None, worker=None, batchable=None,
+                 batch_id=None, retry_policy_dict={}, **kwargs):
         """
         * add task identified by task_id if it doesn't exist
         * if deps is not None, update dependency list
@@ -591,8 +659,10 @@ class CentralPlannerScheduler(Scheduler):
         * add additional workers/stakeholders
         * update priority when needed
         """
-        worker_id = kwargs['worker']
+        assert worker is not None
+        worker_id = worker
         worker_enabled = self.update(worker_id)
+        retry_policy = self._generate_retry_policy(retry_policy_dict)
 
         if worker_enabled:
             _default_task = self._make_task(
@@ -615,16 +685,30 @@ class CentralPlannerScheduler(Scheduler):
         if not task.params:
             task.params = _get_default(params, {})
 
+        if batch_id is not None:
+            task.batch_id = batch_id
+        if status == RUNNING and not task.worker_running:
+            task.worker_running = worker_id
+
         if tracking_url is not None or task.status != RUNNING:
             task.tracking_url = tracking_url
+            if task.batch_id is not None:
+                for batch_task in self._state.get_batch_running_tasks(task.batch_id):
+                    batch_task.tracking_url = tracking_url
+
+        if batchable is not None:
+            task.batchable = batchable
 
         if task.remove is not None:
             task.remove = None  # unmark task for removal so it isn't removed after being added
 
         if expl is not None:
             task.expl = expl
+            if task.batch_id is not None:
+                for batch_task in self._state.get_batch_running_tasks(task.batch_id):
+                    batch_task.expl = expl
 
-        if not (task.status == RUNNING and status == PENDING) or new_deps:
+        if not (task.status in (RUNNING, BATCH_RUNNING) and status == PENDING) or new_deps:
             # don't allow re-scheduling of task while it is running, it must either fail or succeed first
             if status == PENDING or status != task.status:
                 # Update the DB only if there was a acctual change, to prevent noise.
@@ -632,8 +716,6 @@ class CentralPlannerScheduler(Scheduler):
                 # (so checking for status != task.status woule lie)
                 self._update_task_history(task, status)
             self._state.set_status(task, PENDING if status == SUSPENDED else status, self._config)
-            if status == FAILED:
-                task.retry = self._retry_time(task, self._config)
 
         if deps is not None:
             task.deps = set(deps)
@@ -655,22 +737,34 @@ class CentralPlannerScheduler(Scheduler):
 
         self._update_priority(task, priority, worker_id)
 
+        # Because some tasks (non-dynamic dependencies) are `_make_task`ed
+        # before we know their retry_policy, we always set it here
+        task.retry_policy = retry_policy
+
         if runnable and status != FAILED and worker_enabled:
             task.workers.add(worker_id)
             self._state.get_worker(worker_id).tasks.add(task)
             task.runnable = runnable
 
+    @rpc_method()
     def add_worker(self, worker, info, **kwargs):
         self._state.get_worker(worker).add_info(info)
 
+    @rpc_method()
     def disable_worker(self, worker):
         self._state.disable_workers({worker})
 
+    @rpc_method()
     def update_resources(self, **resources):
         if self._resources is None:
             self._resources = {}
         self._resources.update(resources)
 
+    def _generate_retry_policy(self, task_retry_policy_dict):
+        retry_policy_dict = self._config._get_retry_policy()._asdict()
+        retry_policy_dict.update({k: v for k, v in six.iteritems(task_retry_policy_dict) if v is not None})
+        return RetryPolicy(**retry_policy_dict)
+
     def _has_resources(self, needed_resources, used_resources):
         if needed_resources is None:
             return True
@@ -708,10 +802,21 @@ class CentralPlannerScheduler(Scheduler):
                 return False
         return True
 
-    def _retry_time(self, task, config):
-        return time.time() + config.retry_delay
+    def _reset_orphaned_batch_running_tasks(self, worker_id):
+        running_batch_ids = {
+            task.batch_id
+            for task in self._state.get_running_tasks()
+            if task.worker_running == worker_id
+        }
+        orphaned_tasks = [
+            task for task in self._state.get_active_tasks(BATCH_RUNNING)
+            if task.worker_running == worker_id and task.batch_id not in running_batch_ids
+        ]
+        for task in orphaned_tasks:
+            self._state.set_status(task, PENDING)
 
-    def get_work(self, host=None, assistant=False, current_tasks=None, **kwargs):
+    @rpc_method(allow_null=False)
+    def get_work(self, host=None, assistant=False, current_tasks=None, worker=None, **kwargs):
         # TODO: remove any expired nodes
 
         # Algo: iterate over all nodes, find the highest priority node no dependencies and available
@@ -728,12 +833,14 @@ class CentralPlannerScheduler(Scheduler):
         if self._config.prune_on_get_work:
             self.prune()
 
-        worker_id = kwargs['worker']
+        assert worker is not None
+        worker_id = worker
         # Return remaining tasks that have no FAILED descendants
         self.update(worker_id, {'host': host}, get_work=True)
         if assistant:
             self.add_worker(worker_id, [('assistant', assistant)])
 
+        batched_params, unbatched_params, batched_tasks, max_batch_size = None, None, [], 1
         best_task = None
         if current_tasks is not None:
             ct_set = set(current_tasks)
@@ -741,6 +848,10 @@ class CentralPlannerScheduler(Scheduler):
                 if task.worker_running == worker_id and task.id not in ct_set:
                     best_task = task
 
+        if current_tasks is not None:
+            # batch running tasks that weren't claimed since the last get_work go back in the pool
+            self._reset_orphaned_batch_running_tasks(worker_id)
+
         locally_pending_tasks = 0
         running_tasks = []
         upstream_table = {}
@@ -781,6 +892,12 @@ class CentralPlannerScheduler(Scheduler):
                     if len(task.workers) == 1 and not assistant:
                         n_unique_pending += 1
 
+            if (best_task and batched_params and task.family == best_task.family and
+                    len(batched_tasks) < max_batch_size and task.is_batchable() and all(
+                    task.params.get(name) == value for name, value in unbatched_params.items())):
+                for name, params in batched_params.items():
+                    params.append(task.params.get(name))
+                batched_tasks.append(task)
             if best_task:
                 continue
 
@@ -792,6 +909,20 @@ class CentralPlannerScheduler(Scheduler):
             if self._schedulable(task) and self._has_resources(task.resources, greedy_resources):
                 if in_workers and self._has_resources(task.resources, used_resources):
                     best_task = task
+                    batch_param_names, max_batch_size = self._state.get_batcher(
+                        worker_id, task.family)
+                    if batch_param_names and task.is_batchable():
+                        try:
+                            batched_params = {
+                                name: [task.params[name]] for name in batch_param_names
+                            }
+                            unbatched_params = {
+                                name: value for name, value in task.params.items()
+                                if name not in batched_params
+                            }
+                            batched_tasks.append(task)
+                        except KeyError:
+                            batched_params, unbatched_params = None, None
                 else:
                     workers = itertools.chain(task.workers, [worker_id]) if assistant else task.workers
                     for task_worker in workers:
@@ -810,7 +941,23 @@ class CentralPlannerScheduler(Scheduler):
                  'task_id': None,
                  'n_unique_pending': n_unique_pending}
 
-        if best_task:
+        if len(batched_tasks) > 1:
+            batch_string = '|'.join(task.id for task in batched_tasks)
+            batch_id = hashlib.md5(batch_string.encode('utf-8')).hexdigest()
+            for task in batched_tasks:
+                self._state.set_batch_running(task, batch_id, worker_id)
+
+            combined_params = best_task.params.copy()
+            combined_params.update(batched_params)
+
+            reply['task_id'] = None
+            reply['task_family'] = best_task.family
+            reply['task_module'] = getattr(best_task, 'module', None)
+            reply['task_params'] = combined_params
+            reply['batch_id'] = batch_id
+            reply['batch_task_ids'] = [task.id for task in batched_tasks]
+
+        elif best_task:
             self._state.set_status(best_task, RUNNING, self._config)
             best_task.worker_running = worker_id
             best_task.time_running = time.time()
@@ -823,6 +970,7 @@ class CentralPlannerScheduler(Scheduler):
 
         return reply
 
+    @rpc_method(attempts=1)
     def ping(self, **kwargs):
         worker_id = kwargs['worker']
         self.update(worker_id)
@@ -849,9 +997,8 @@ class CentralPlannerScheduler(Scheduler):
                     elif upstream_status_table[dep_id] == '' and dep.deps:
                         # This is the postorder update step when we set the
                         # status based on the previously calculated child elements
-                        status = max((upstream_status_table.get(a_task_id, '')
-                                      for a_task_id in dep.deps),
-                                     key=UPSTREAM_SEVERITY_KEY)
+                        upstream_severities = list(upstream_status_table.get(a_task_id) for a_task_id in dep.deps if a_task_id in upstream_status_table) or ['']
+                        status = min(upstream_severities, key=UPSTREAM_SEVERITY_KEY)
                         upstream_status_table[dep_id] = status
             return upstream_status_table[dep_id]
 
@@ -870,6 +1017,7 @@ class CentralPlannerScheduler(Scheduler):
             'priority': task.priority,
             'resources': task.resources,
             'tracking_url': getattr(task, "tracking_url", None),
+            'status_message': getattr(task, "status_message", None)
         }
         if task.status == DISABLED:
             ret['re_enable_able'] = task.scheduler_disable_time is not None
@@ -877,6 +1025,7 @@ class CentralPlannerScheduler(Scheduler):
             ret['deps'] = list(task.deps if deps is None else deps)
         return ret
 
+    @rpc_method()
     def graph(self, **kwargs):
         self.prune()
         serialized = {}
@@ -918,7 +1067,7 @@ class CentralPlannerScheduler(Scheduler):
 
             task = self._state.get_task(task_id)
             if task is None or not task.family:
-                logger.warn('Missing task for id [%s]', task_id)
+                logger.debug('Missing task for id [%s]', task_id)
 
                 # NOTE : If a dependency is missing from self._state there is no way to deduce the
                 #        task family and parameters.
@@ -952,12 +1101,14 @@ class CentralPlannerScheduler(Scheduler):
 
         return serialized
 
+    @rpc_method()
     def dep_graph(self, task_id, include_done=True, **kwargs):
         self.prune()
         if not self._state.has_task(task_id):
             return {}
         return self._traverse_graph(task_id, include_done=include_done)
 
+    @rpc_method()
     def inverse_dep_graph(self, task_id, include_done=True, **kwargs):
         self.prune()
         if not self._state.has_task(task_id):
@@ -969,7 +1120,8 @@ class CentralPlannerScheduler(Scheduler):
         return self._traverse_graph(
             task_id, dep_func=lambda t: inverse_graph[t.id], include_done=include_done)
 
-    def task_list(self, status, upstream_status, limit=True, search=None, **kwargs):
+    @rpc_method()
+    def task_list(self, status='', upstream_status='', limit=True, search=None, **kwargs):
         """
         Query for a subset of tasks by status.
         """
@@ -985,8 +1137,7 @@ class CentralPlannerScheduler(Scheduler):
             def filter_func(t):
                 return all(term in t.pretty_id for term in terms)
         for task in filter(filter_func, self._state.get_active_tasks(status)):
-            if (task.status != PENDING or not upstream_status or
-                    upstream_status == self._upstream_status(task.id, upstream_status_table)):
+            if task.status != PENDING or not upstream_status or upstream_status == self._upstream_status(task.id, upstream_status_table):
                 serialized = self._serialize_task(task.id, False)
                 result[task.id] = serialized
         if limit and len(result) > self._config.max_shown_tasks:
@@ -1000,6 +1151,7 @@ class CentralPlannerScheduler(Scheduler):
         else:
             return task_id
 
+    @rpc_method()
     def worker_list(self, include_running=True, **kwargs):
         self.prune()
         workers = [
@@ -1031,6 +1183,7 @@ class CentralPlannerScheduler(Scheduler):
                 worker['running'] = tasks
         return workers
 
+    @rpc_method()
     def resource_list(self):
         """
         Resources usage info and their consumers (tasks).
@@ -1066,6 +1219,7 @@ class CentralPlannerScheduler(Scheduler):
                 ret[resource]['used'] = 0
         return ret
 
+    @rpc_method()
     def task_search(self, task_str, **kwargs):
         """
         Query for a subset of tasks by task_id.
@@ -1081,6 +1235,7 @@ class CentralPlannerScheduler(Scheduler):
                 result[task.status][task.id] = serialized
         return result
 
+    @rpc_method()
     def re_enable_task(self, task_id):
         serialized = {}
         task = self._state.get_task(task_id)
@@ -1089,6 +1244,7 @@ class CentralPlannerScheduler(Scheduler):
             serialized = self._serialize_task(task_id)
         return serialized
 
+    @rpc_method()
     def fetch_error(self, task_id, **kwargs):
         if self._state.has_task(task_id):
             task = self._state.get_task(task_id)
@@ -1096,6 +1252,23 @@ class CentralPlannerScheduler(Scheduler):
         else:
             return {"taskId": task_id, "error": ""}
 
+    @rpc_method()
+    def set_task_status_message(self, task_id, status_message):
+        if self._state.has_task(task_id):
+            task = self._state.get_task(task_id)
+            task.status_message = status_message
+            if task.status == RUNNING and task.batch_id is not None:
+                for batch_task in self._state.get_batch_running_tasks(task.batch_id):
+                    batch_task.status_message = status_message
+
+    @rpc_method()
+    def get_task_status_message(self, task_id):
+        if self._state.has_task(task_id):
+            task = self._state.get_task(task_id)
+            return {"taskId": task_id, "statusMessage": task.status_message}
+        else:
+            return {"taskId": task_id, "statusMessage": ""}
+
     def _update_task_history(self, task, status, host=None):
         try:
             if status == DONE or status == FAILED:
