@@ -14,6 +14,7 @@ function visualiserApp(luigi) {
     var taskIcons = {
         PENDING: 'pause',
         RUNNING: 'play',
+        BATCH_RUNNING: 'play',
         DONE: 'check',
         FAILED: 'times',
         UPSTREAM_FAILED: 'warning',
@@ -84,7 +85,8 @@ function visualiserApp(luigi) {
             status: task.status,
             graph: (task.status == "PENDING" || task.status == "RUNNING" || task.status == "DONE"),
             error: task.status == "FAILED",
-            re_enable: task.status == "DISABLED" && task.re_enable_able
+            re_enable: task.status == "DISABLED" && task.re_enable_able,
+            statusMessage: task.status_message
         };
     }
 
@@ -101,6 +103,10 @@ function visualiserApp(luigi) {
                 iconClass = 'fa-play';
                 iconColor = 'aqua';
                 break;
+            case 'BATCH_RUNNING':
+                iconClass = 'fa-play';
+                iconColor = 'purple';
+                break;
             case 'DONE':
                 iconClass = 'fa-check';
                 iconColor = 'green';
@@ -246,6 +252,7 @@ function visualiserApp(luigi) {
         resource.percent_used = 100 * resource.num_used / resource.num_total;
         if (resource.percent_used >= 100) {
             resource.bar_type = 'danger';
+            resource.percent_used = 100;
         } else if (resource.percent_used > 50) {
             resource.bar_type = 'warning';
         } else {
@@ -289,6 +296,16 @@ function visualiserApp(luigi) {
         $("#errorModal").modal({});
     }
 
+    function showStatusMessage(data) {
+        $("#statusMessageModal").empty().append(renderTemplate("statusMessageTemplate", data));
+        $("#statusMessageModal .refresh").on('click', function() {
+            luigi.getTaskStatusMessage(data.taskId, function(data) {
+                $("#statusMessageModal pre").html(data.statusMessage);
+            });
+        }).trigger('click');
+        $("#statusMessageModal").modal({});
+    }
+
     function preProcessGraph(dependencyGraph) {
         var extraNodes = [];
         var seen = {};
@@ -619,6 +636,7 @@ function visualiserApp(luigi) {
         var mostImportantCategory = function (cat1, cat2) {
             var priorities = [
                 'RUNNING',
+                'BATCH_RUNNING',
                 'DONE',
                 'PENDING',
                 'UPSTREAM_DISABLED',
@@ -722,31 +740,35 @@ function visualiserApp(luigi) {
             updateTaskCategory(dt, 'RUNNING', runningTasks);
         });
 
-        var ajax2 = luigi.getFailedTaskList(function(failedTasks) {
+        var ajax2 = luigi.getBatchRunningTaskList(function(batchRunningTasks) {
+            updateTaskCategory(dt, 'BATCH_RUNNING', batchRunningTasks);
+        });
+
+        var ajax3 = luigi.getFailedTaskList(function(failedTasks) {
             updateTaskCategory(dt, 'FAILED', failedTasks);
         });
 
-        var ajax3 = luigi.getUpstreamFailedTaskList(function(upstreamFailedTasks) {
+        var ajax4 = luigi.getUpstreamFailedTaskList(function(upstreamFailedTasks) {
             updateTaskCategory(dt, 'UPSTREAM_FAILED', upstreamFailedTasks);
         });
 
-        var ajax4 = luigi.getDisabledTaskList(function(disabledTasks) {
+        var ajax5 = luigi.getDisabledTaskList(function(disabledTasks) {
             updateTaskCategory(dt, 'DISABLED', disabledTasks);
         });
 
-        var ajax5 = luigi.getUpstreamDisabledTaskList(function(upstreamDisabledTasks) {
+        var ajax6 = luigi.getUpstreamDisabledTaskList(function(upstreamDisabledTasks) {
             updateTaskCategory(dt, 'UPSTREAM_DISABLED', upstreamDisabledTasks);
         });
 
-        var ajax6 = luigi.getPendingTaskList(function(pendingTasks) {
+        var ajax7 = luigi.getPendingTaskList(function(pendingTasks) {
             updateTaskCategory(dt, 'PENDING', pendingTasks);
         });
 
-        var ajax7 = luigi.getDoneTaskList(function(doneTasks) {
+        var ajax8 = luigi.getDoneTaskList(function(doneTasks) {
             updateTaskCategory(dt, 'DONE', doneTasks);
         });
 
-        $.when(ajax1, ajax2, ajax3, ajax4, ajax5, ajax6, ajax7).done(function () {
+        $.when(ajax1, ajax2, ajax3, ajax4, ajax5, ajax6, ajax7, ajax8).done(function () {
             dt.draw();
 
             $('.sidebar').html(renderSidebar(dt.column(1).data()));
@@ -815,6 +837,11 @@ function visualiserApp(luigi) {
 
         luigi.getWorkerList(function(workers) {
             $("#workerList").append(renderWorkers(workers));
+
+            $('.worker-table tbody').on('click', 'td .statusMessage', function() {
+                var data = $(this).data();
+                showStatusMessage(data);
+            });
         });
 
         luigi.getResourceList(function(resources) {
@@ -913,6 +940,11 @@ function visualiserApp(luigi) {
             });
         });
 
+        $('#taskTable tbody').on('click', 'td.details-control .statusMessage', function () {
+            var data = $(this).data();
+            showStatusMessage(data);
+        });
+
         $('.navbar-nav').on('click', 'a', function () {
             var tabName = $(this).data('tab');
             updateSidebar(tabName);
